/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class KaHIPWrapper */

#ifndef _Included_KaHIPWrapper
#define _Included_KaHIPWrapper
#ifdef __cplusplus
extern "C" {
#endif
#undef KaHIPWrapper_KAHIP_FAST
#define KaHIPWrapper_KAHIP_FAST 0L
#undef KaHIPWrapper_KAHIP_ECO
#define KaHIPWrapper_KAHIP_ECO 1L
#undef KaHIPWrapper_KAHIP_STRONG
#define KaHIPWrapper_KAHIP_STRONG 2L
#undef KaHIPWrapper_KAHIP_FASTSOCIAL
#define KaHIPWrapper_KAHIP_FASTSOCIAL 3L
#undef KaHIPWrapper_KAHIP_ECOSOCIAL
#define KaHIPWrapper_KAHIP_ECOSOCIAL 4L
#undef KaHIPWrapper_KAHIP_STRONGSOCIAL
#define KaHIPWrapper_KAHIP_STRONGSOCIAL 5L
/*
 * Class:     KaHIPWrapper
 * Method:    cnativeKaffpa
 * Signature: (I[I[I[I[IIDZIILKaHIPWrapperResult;)V
 */
JNIEXPORT void JNICALL Java_KaHIPWrapper_cnativeKaffpa
  (JNIEnv *, jclass, jint, jintArray, jintArray, jintArray, jintArray, jint, jdouble, jboolean, jint, jint, jobject);

#ifdef __cplusplus
}
#endif
#endif
